from pyodm.utils.path_utils import PathUtils


def odm_xsd_description():
    base = PathUtils.folder("pyodm", __file__).joinpath("model", "v2", "resources", "schema")
    xsd_files = ["ODM-admindata.xsd", "ODM-clinicaldata.xsd", "ODM-foundation.xsd",
                 "ODM-protocol.xsd", "ODM-referencedata.xsd", "ODM-study.xsd"]
    return [base.joinpath(xsd) for xsd in xsd_files]


def odm_specification_description():
    base = PathUtils.folder("pyodm", __file__).joinpath("resources")
    files = ["SpecificationV2.xml"]
    return [base.joinpath(file) for file in files]



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class ODM(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/ODM
    """
    
    FileType = Model.Attribute()
    
    Granularity = Model.Attribute()
    
    Context = Model.Attribute()
    
    FileOID = Model.Attribute()
    
    CreationDateTime = Model.Attribute()
    
    PriorFileOID = Model.Attribute()
    
    AsOfDateTime = Model.Attribute()
    
    ODMVersion = Model.Attribute()
    
    Originator = Model.Attribute()
    
    SourceSystem = Model.Attribute()
    
    SourceSystemVersion = Model.Attribute()
    
    
    Description = Model.ManyElements()
    
    Study = Model.ManyElements()
    
    AdminData = Model.ManyElements()
    
    ReferenceData = Model.ManyElements()
    
    ClinicalData = Model.ManyElements()
    
    Association = Model.ManyElements()
    



import pyodm.model.definition as Model
# from pyodm.core.parser import Parser


class EntityMeta(type):
    def __init__(cls, name, bases, attr_dict):
        super().__init__(name, bases, attr_dict)
        for key, attr in attr_dict.items():
            if isinstance(attr, Model.Attribute):
                attr.set_name(key)
            if isinstance(attr, Model.OneElement):
                attr.set_name(key)
            if isinstance(attr, Model.ManyElements):
                attr.set_name(key)





from pyodm.utils.search.odm_dfs import ODMDfs
from pyodm.utils.odm_utils import OdmUtils


class Stream:
    def __init__(self, data):
        self.dfs = self.dfs_clazz()(data)

    def find(self, _lambda):
        return Optional(self.dfs.find(_lambda=_lambda))

    def dfs_clazz(self): return ODMDfs


class Optional:
    def __init__(self, data):
        self._data = data

    def get(self):
        return self._data

    def find(self, _lambda):
        return self.as_stream().find(_lambda=_lambda)

    def update(self, **attributes):
        """
        更新 ODM 的属性
        :param attributes:
        :type attributes:
        :return:
        :rtype:
        """
        for name, value in attributes.items():
            self._data.__dict__.get(name).set_value(value)
        return self

    def update_text(self, value):
        """
        更新 ODM 的文本值
        :param value:
        :type value:
        :return:
        :rtype:
        """
        self._data.set_value(value)
        return self

    def merge(self, odm_element):
        """
        合并 两个 ODM ELEMENT
        前提是 被合并对象 odm_element 与 合并对象的 根节点一致
        OneElement ,以当前元素为基准，合并属性，并且 合并子元素
        ManyElement , 在原有的元素上，添加新的元素
        :param odm_element:
        :type odm_element:
        :return:
        :rtype:
        """
        if self._data.get_name() != odm_element.get_name():
            raise Exception(f'{self._data.get_name()} is different {odm_element.get_name()}')
        OdmUtils.merge(self._data, odm_element)

    def replace(self, name, odm_element):
        """
        替换指定的子元素
        :param name: element name
        :type name: str
        :param odm_element: 替换的元素
        :type odm_element:
        :return:
        :rtype:
        """
        if self._data.__dict__.get(name) is None: raise Exception(f'{self._data.get_name()} have not element {name}')
        replaced = self._data.__dict__.get(name)
        """处理 OneElement 和 实例 直接替换"""
        if OdmUtils.is_entity(replaced) or OdmUtils.is_one_element(replaced):
            setattr(self._data, name, odm_element)
        """处理ManyElements"""
        if OdmUtils.is_many_elements(replaced):
            """如果 为空，则直接增加一个项 """
            if OdmUtils.useless(self._data, name):
                self.append(name, odm_element)
            else:
                """如果 不为空，则查找相似的替换 """
                for i in range(len(replaced.array)):
                    if OdmUtils.like(replaced.index(i), odm_element):
                        replaced.array[i] = odm_element
                        break

    def append(self, name, odm_element):
        """
        ManyElements 添加一个元素
        :return:
        :rtype:
        """
        if OdmUtils.is_many_elements(self._data.__dict__.get(name)):
            if OdmUtils.useless(self._data, name):  self._data.__dict__.get(name).set_name(name)
            OdmUtils.append(self._data.__dict__.get(name), odm_element)
        else:
            raise Exception(f"{name} is not ManyElements")

    def add(self, element_entity):
        """
        增加一个 OneElement 实例
        前提是：
        1. 新增的 OneElement 存在
        2. OneElement 还没有被实例化
        :param element_entity:  Element 的实例
        :type element_entity:
        :return:
        :rtype:
        """
        if not OdmUtils.is_one_element(self._data.__dict__.get(element_entity.get_name())):
            raise Exception(f'{element_entity.get_name()} is not OneElement')
        OdmUtils.add(self._data, element_entity)


    def as_stream(self):
        return Stream(self._data)




import pytest

from pyodm.core.source.path_source import PathSource
from pyodm.factory.cdisc_xml_specification_factory import CdiscXMLSpecificationFactory
from pyodm.tests.test_cdisc_factory.common_assert import assert_case2, assert_case6, assert_case5, assert_case4, \
    assert_case3


@pytest.fixture
def specification_file():
    return "D:\\github\\pyodm\\src\\pyodm\\resources\\SpecificationV2.xml"


def test_case2(specification_file):
    data_file = PathSource("resources", "data", "odm_template.xml")
    cdisc = CdiscXMLSpecificationFactory(data_file=data_file, specification_files=[specification_file])
    assert_case2(cdisc)


def test_case3(specification_file):
    data_file = PathSource("resources", "data", "test_case2.xml")
    cdisc = CdiscXMLSpecificationFactory(data_file=data_file, specification_files=[specification_file])
    assert_case3(cdisc)


def test_case4(specification_file):
    data_file = PathSource("resources", "data", "transaction_history.xml")
    cdisc = CdiscXMLSpecificationFactory(data_file=data_file, specification_files=[specification_file])
    assert_case4(cdisc)


def test_case5(specification_file):
    data_file = PathSource("resources", "data", "transaction_history.xml")
    cdisc = CdiscXMLSpecificationFactory(data_file=data_file, specification_files=[specification_file])
    assert_case5(cdisc)


def test_case6(specification_file):
    data_file = PathSource("resources", "data", "odm_testcase5.xml")
    cdisc = CdiscXMLSpecificationFactory(data_file=data_file, specification_files=[specification_file])
    assert_case6(cdisc)



import pytest
import lxml.etree as Etree

from pyodm.utils.validator import Validator

def test_xml_xsd_valid():
    assert Validator.odm("resources/odm_template.xml") == True


def test_xml_xsd_unvalid():
    with pytest.raises(Etree.DocumentInvalid):
        Validator.odm("resources/un_odm_xml.xml")



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class ItemGroupDef(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/ItemGroupDef
    """
    
    OID = Model.Attribute()
    
    Name = Model.Attribute()
    
    Repeating = Model.Attribute()
    
    RepeatingLimit = Model.Attribute()
    
    IsReferenceData = Model.Attribute()
    
    Structure = Model.Attribute()
    
    ArchiveLocationID = Model.Attribute()
    
    DatasetName = Model.Attribute()
    
    Domain = Model.Attribute()
    
    Type = Model.Attribute()
    
    Purpose = Model.Attribute()
    
    StandardOID = Model.Attribute()
    
    IsNonStandard = Model.Attribute()
    
    HasNoData = Model.Attribute()
    
    CommentOID = Model.Attribute()
    
    
    Description = Model.OneElement()
    
    Class = Model.ManyElements()
    
    Coding = Model.ManyElements()
    
    WorkflowRef = Model.ManyElements()
    
    Origin = Model.ManyElements()
    
    Alias = Model.ManyElements()
    
    Leaf = Model.ManyElements()
    
    ItemGroupRef = Model.OneElement()
    
    ItemRef = Model.OneElement()
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class TranslatedText(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/TranslatedText
    """
    
    xml_lang = Model.Attribute()
    
    Type = Model.Attribute()
    
    
    xhtml_div = Model.ManyElements()
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class ClinicalData(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/ClinicalData+Element
    """
    
    StudyOID = Model.Attribute()
    
    MetaDataVersionOID = Model.Attribute()
    
    
    SubjectData = Model.ManyElements()
    
    ItemGroupData = Model.ManyElements()
    
    Query = Model.ManyElements()
    
    AuditRecord = Model.OneElement()
    
    Signature = Model.OneElement()
    
    Annotation = Model.ManyElements()
    



from pyodm.exceptions import CdiscFactoryException
from pyodm.factory.xpath.xpath_tree import XpathTree


class CdiscRegistry:
    """
    注册 ODM 元素的 Class 类型
    """
    def __init__(self):
        self._registry_classes = dict()
        self.definition_tree = XpathTree()

    def get(self, name: str):
        """
        根据 ODM 的元素名, 获取 ODM Class
        :param name: ODM 元素名
        :type name:
        :return:
        :rtype:
        """
        if self._registry_classes.get(name) is None:
            raise CdiscFactoryException(f"{name} No Registry Class")
        return self._registry_classes.get(name)

    def registry_cdisc(self, name, cdisc_class):
        """
        根据 ODM 元素名, 注册 ODM Class
        :param name:
        :type name:
        :param cdisc_class:
        :type cdisc_class:
        :return:
        :rtype:
        """
        self._registry_classes[name] = cdisc_class



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Study(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Study
    """
    
    OID = Model.Attribute()
    
    StudyName = Model.Attribute()
    
    ProtocolName = Model.Attribute()
    
    VersionID = Model.Attribute()
    
    VersionName = Model.Attribute()
    
    Status = Model.Attribute()
    
    
    Description = Model.OneElement()
    
    MetaDataVersion = Model.ManyElements()
    



import abc
import copy


class DFS(abc.ABC):
    def __init__(self, data, cached: bool = False):
        self._data = data
        self._stack = []
        self._cached = cached
        self._scan(self._data)

    @abc.abstractmethod
    def neighbor_node(self, node) -> list:
        ...

    @abc.abstractmethod
    def pruning_strategy(self, node) -> bool:
        ...

    def _scan(self, node):
        if self.pruning_strategy(node):
            self._stack.append(node)
            for child in self.neighbor_node(node):
                self._scan(child)

    def find(self, _lambda):
        __stack = self._stack if not self._cached else copy.deepcopy(self._stack)
        while __stack:
            element = __stack.pop()
            if _lambda(element):
                self.clear_stack()
                return element
        self.clear_stack()
        return None

    def clear_stack(self):
        if not self._cached:
            self._stack = []



from pyodm.model.meta.entity_meta import EntityMeta
import pyodm.model.definition as Model
#


class CdiscODMEntity(metaclass=EntityMeta):
    """
    自定义ODM的元素的时候，必须继承的元类
    """
    def __init__(self):
        self._name = ""
        self._value = None
        for name, instance in vars(self.__class__).items():
            if isinstance(instance, Model.Attribute) \
                    or isinstance(instance, Model.OneElement) \
                    or isinstance(instance, Model.ManyElements):
                setattr(self, name, instance.__class__())

    def get_name(self):
        return self._name

    def get_value(self): return self._value

    def set_value(self, value): self._value = value

    def set_name(self, value):
        self._name = value

    def is_blank(self) -> bool: return self.get_value() is None

    def as_stream(self):
        from pyodm.utils.stream import Stream
        return Stream(self)



from typing import Type


class Mixin:

    @classmethod
    def first(cls): return cls

    @classmethod
    def find(cls, **attributes): return cls

    @classmethod
    def index(cls, i): return cls

    @classmethod
    def array(cls): return [cls]

    @property
    def name(self): ...

    @property
    def value(self): ...


class Attribute:
    @property
    def name(self): ...

    @property
    def value(self): ...


class Model:
    @staticmethod
    def Attribute() -> Attribute: ...

    @staticmethod
    def OneElement() -> Mixin: ...


class SignatureRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SignatureRef+Element
    """

    SignatureOID = Model.Attribute()


class Signature(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Signature+Element
    """
    SignatureRef: SignatureRef = Model.OneElement()


class ClinicalData(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ClinicalData+Element
    """

    Signature: Signature = Model.OneElement()



import copy

from lxml import etree

from pyodm.core.support.abstract_configuration_reader import AbstractConfigurationReader
import pyodm.model.definition as Model
from pyodm.factory.cdisc_registry import CdiscRegistry
from pyodm.factory.xpath.xpath_node import XpathNode
from pyodm.model.meta.cdisc_odm_entity import CdiscODMEntity


# TODO 变量的域的区分 xmlns(基本功能完成后再做)
class CdiscConfigurationXsdReader(AbstractConfigurationReader):
    """
    读取 ODM XSD 格式的配置文件
    https://github.com/cdisc-org/DataExchange-ODM/tree/main/schema
    """
    def __init__(self):
        self.attributeGroups = dict()
        self.complexTypes = dict()
        self.cdiscElements = dict()
        self.groups = dict()

    def load(self, xsd_file):
        xsd_tree = etree.parse(xsd_file)
        root = xsd_tree.getroot()
        for element in root:
            if not isinstance(element.tag, str): continue
            name = etree.QName(element.tag).localname
            if name == "attributeGroup":
                self.attributeGroup(element)
            if name == "complexType":
                self.complexType(element)
            if name == "element":
                self.cdisc_element(element)
            if name == 'group':
                self.group(element)
            # print(element)

    def group(self, element):
        self.groups[element.attrib.get("name")] = dict(elements=[])
        for x in element:
            if not isinstance(x.tag, str): continue
            if etree.QName(x.tag).localname == "sequence":
                for y in x:
                    if not isinstance(y.tag, str): continue
                    if etree.QName(y.tag).localname == "element":
                        ref = y.attrib.get("ref").replace(":", "_")
                        cls_type = "model.OneElement()" if y.attrib.get("maxOccurs") is None else "model.ManyElements()"
                        self.groups[element.attrib.get("name")]["elements"].append(
                            dict(name=ref, cls_type=cls_type))

    def cdisc_element(self, element):
        self.cdiscElements[element.attrib.get("name")] = element.attrib.get("type")

    def attributeGroup(self, element):
        self.attributeGroups[element.attrib.get("name")] = dict()
        for attribute in element:
            if not isinstance(attribute.tag, str): continue
            if etree.QName(attribute.tag).localname == "attribute":
                name = attribute.attrib.get("name") or attribute.attrib.get("ref").replace(":", "_")
                self.attributeGroups[element.attrib.get("name")].update({
                    name: "model.Attribute()"})

    def complexType(self, element):
        """
        临时的解析代码
        :param element:
        :type element:
        :return:
        :rtype:
        """
        self.complexTypes[element.attrib.get("name")] = dict(elements=[], attributes=[], elementGroup=[])
        for x in element:
            if not isinstance(x.tag, str): continue
            if etree.QName(x.tag).localname == "sequence":
                for y in x:
                    if not isinstance(y.tag, str): continue
                    if etree.QName(y.tag).localname == "element":
                        ref = y.attrib.get("ref").replace(":", "_")
                        cls_type = "model.OneElement()" if y.attrib.get("maxOccurs") is None else "model.ManyElements()"
                        self.complexTypes[element.attrib.get("name")]["elements"].append(
                            dict(name=ref, cls_type=cls_type))
                    if etree.QName(y.tag).localname == "group":
                        ref = y.attrib.get("ref").replace(":", "_")
                        self.complexTypes[element.attrib.get("name")]["elementGroup"].append(ref)
            if etree.QName(x.tag).localname == "attributeGroup":
                self.complexTypes[element.attrib.get("name")]["attributes"].append(
                    x.attrib.get("ref").replace(":", "_"))
            if etree.QName(x.tag).localname == "simpleContent":
                for y in x:
                    if not isinstance(y.tag, str): continue
                    if etree.QName(y.tag).localname == "extension":
                        for z in y:
                            if not isinstance(z.tag, str): continue
                            if etree.QName(z.tag).localname == "attributeGroup":
                                if not isinstance(z.tag, str): continue
                                self.complexTypes[element.attrib.get("name")]["attributes"].append(
                                    z.attrib.get("ref").replace(":", "_"))

    def cdisc(self):
        results = {}
        for name, complex_type in self.cdiscElements.items():
            results[name] = {"elements": copy.deepcopy(self.complexTypes[complex_type].get("elements")),
                             "attributes": {}}
            for groupName in self.complexTypes[complex_type].get("elementGroup"):
                if self.groups.get(groupName) is None: continue
                for element in self.groups[groupName].get("elements"):
                    results[name]["elements"].append(element)
            for ref in self.complexTypes[complex_type].get("attributes"):
                attrib = self.attributeGroups.get(ref)
                if attrib is not None:
                    results[name]["attributes"].update(attrib)
        return results

    def load_cdisc_definition(self, registry: CdiscRegistry, xsd_files: list):
        for xsd_file in xsd_files:
            self.load(xsd_file=xsd_file)
        for name, clazz_info in self.cdisc().items():
            registry.registry_cdisc(name=name, cdisc_class=self._clazz_parser(name, clazz_info, registry))

    def _clazz_parser(self, name, clazz_info: dict, registry: CdiscRegistry):
        __definition = {
            "model.Attribute()": Model.Attribute,
            "model.OneElement()": Model.OneElement,
            "model.ManyElements()": Model.ManyElements
        }
        attrs = {}
        for element in clazz_info.get("elements"):
            cdisc_definition = __definition.get(element.get("cls_type"))()
            cdisc_definition.parent = name
            attrs[element.get("name")] = cdisc_definition
        for attribute_name, cls_type in clazz_info.get("attributes").items():
            attrs[attribute_name] = __definition.get(cls_type)()
        registry.definition_tree.add(XpathNode(name), [XpathNode(e.get('name')) for e in clazz_info.get("elements")])
        return type(name, (CdiscODMEntity,), attrs)





from pyodm.core.support.abstract_data_loader import AbstractDataLoader
from pyodm.core.support.abstract_data_reader import AbstractDataReader
from pyodm.core.xml.reader.configuration.cdisc_configuration_xsd_reader import CdiscConfigurationXsdReader
from pyodm.factory.abstract_hierarchy_factory import AbstractHierarchyFactory
from pyodm.core.json.reader.json_data_reader import JsonDataReader


class CdiscJSONFactory(AbstractHierarchyFactory):
    def __init__(self, data_loader: AbstractDataLoader, configuration_files):
        super().__init__(configuration_files=configuration_files)
        self.data_loader = data_loader

    def clazz_reader(self):
        CdiscConfigurationXsdReader().load_cdisc_definition(self, self._configuration_files)

    def data_reader(self) -> AbstractDataReader:
        return JsonDataReader(self)



from pyodm.utils.odm_utils import OdmUtils
from pyodm.utils.search.dfs import DFS
import pyodm.model.definition as Model


class ODMDfs(DFS):

    def pruning_strategy(self, node) -> bool:
        return True

    def __init__(self, data, cached=False):
        super().__init__(data, cached)

    def neighbor_node(self, node):
        __neighbor = []
        for element in vars(node).values():
            if isinstance(element, Model.OneElement):
                __neighbor.append(element)
            elif isinstance(element, Model.ManyElements):
                for i in element.array:
                    __neighbor.append(i)
            elif OdmUtils.is_entity(element):
                __neighbor.append(element)
        return __neighbor



import copy
import os

from jinja2 import FileSystemLoader, Environment
from lxml import etree


class GeneratePy:
    def __init__(self, xsd_file):
        self.xsd_file = xsd_file
        self.attributeGroups = dict()
        self.complexTypes = dict()
        self.cdiscElements = dict()
        self.groups = dict()

    def generate(self):
        xsd_tree = etree.parse(self.xsd_file)
        root = xsd_tree.getroot()
        for element in root:
            if not isinstance(element.tag, str): continue
            name = etree.QName(element.tag).localname
            if name == "attributeGroup":
                self.attributeGroup(element)
            if name == "complexType":
                self.complexType(element)
            if name == "element":
                self.cdisc_element(element)
            if name == 'group':
                self.group(element)
            # print(element)

    def group(self, element):
        self.groups[element.attrib._get("name")] = dict(elements=[])
        for x in element:
            if not isinstance(x.tag, str): continue
            if etree.QName(x.tag).localname == "sequence":
                for y in x:
                    if not isinstance(y.tag, str): continue
                    if etree.QName(y.tag).localname == "element":
                        ref = y.attrib._get("ref").replace(":", "_")
                        cls_type = "model.OneElement()" if y.attrib._get("maxOccurs") is None else "model.ManyElements()"
                        self.groups[element.attrib._get("name")]["elements"].append(
                            dict(name=ref, cls_type=cls_type))

    def cdisc_element(self, element):
        self.cdiscElements[element.attrib._get("name")] = element.attrib._get("type")

    def attributeGroup(self, element):
        self.attributeGroups[element.attrib._get("name")] = dict()
        for attribute in element:
            if not isinstance(attribute.tag, str): continue
            if etree.QName(attribute.tag).localname == "attribute":
                name = attribute.attrib._get("name") or attribute.attrib._get("ref").replace(":", "_")
                self.attributeGroups[element.attrib._get("name")].update({
                    name: "model.Attribute()"})

    def complexType(self, element):
        """
        临时的解析代码
        :param element:
        :type element:
        :return:
        :rtype:
        """
        self.complexTypes[element.attrib._get("name")] = dict(elements=[], attributes=[], elementGroup=[])
        for x in element:
            if not isinstance(x.tag, str): continue
            if etree.QName(x.tag).localname == "sequence":
                for y in x:
                    if not isinstance(y.tag, str): continue
                    if etree.QName(y.tag).localname == "element":
                        ref = y.attrib._get("ref").replace(":", "_")
                        cls_type = "model.OneElement()" if y.attrib._get("maxOccurs") is None else "model.ManyElements()"
                        self.complexTypes[element.attrib._get("name")]["elements"].append(dict(name=ref, cls_type=cls_type))
                    if etree.QName(y.tag).localname == "group":
                        ref = y.attrib._get("ref").replace(":", "_")
                        self.complexTypes[element.attrib._get("name")]["elementGroup"].append(ref)
            if etree.QName(x.tag).localname == "attributeGroup":
                self.complexTypes[element.attrib._get("name")]["attributes"].append(x.attrib._get("ref").replace(":", "_"))
            if etree.QName(x.tag).localname == "simpleContent":
                for y in x:
                    if not isinstance(y.tag, str): continue
                    if etree.QName(y.tag).localname == "extension":
                        for z in y:
                            if not isinstance(z.tag, str): continue
                            if etree.QName(z.tag).localname == "attributeGroup":
                                if not isinstance(z.tag, str): continue
                                self.complexTypes[element.attrib._get("name")]["attributes"].append(
                                            z.attrib._get("ref").replace(":", "_"))


    def console(self):
        print(self.cdisc())
        # print(self.attributeGroups)
        # print(self.cdiscElements)
        # print(self.complexTypes)

    def cdisc(self):
        results = {}
        for name, complex_type in self.cdiscElements.items():
            results[name] = {"elements": copy.deepcopy(self.complexTypes[complex_type]._get("elements")), "attributes": {}}
            for groupName in self.complexTypes[complex_type]._get("elementGroup"):
                if self.groups.get(groupName) is None: continue
                for element in self.groups[groupName]._get("elements"):
                    results[name]["elements"].append(element)
            for ref in self.complexTypes[complex_type]._get("attributes"):
                attrib = self.attributeGroups.get(ref)
                if attrib is not None:
                    results[name]["attributes"].update(attrib)
        return results

    def write_spec(self, cdiscs):
        file_loader = FileSystemLoader('./')
        env = Environment(loader=file_loader)
        template = env.get_template("template/spec.template")
        output = template.render(cdiscs=cdiscs)
        if not os.path.exists('out_put'): os.mkdir('out_put')
        with open(os.path.join('out_put', 'spec.py'), "w", encoding="UTF-8") as f:
            f.write(output)

    def write_cdisc(self, cdiscs):
        file_loader = FileSystemLoader('./')
        env = Environment(loader=file_loader)
        template = env.get_template("template/cdsic.template")

        if not os.path.exists('out_put'): os.mkdir('out_put')
        for cdiscName, cdisc in cdiscs.items():
            output = template.render(cdiscName=cdiscName, cdisc=cdisc)
            with open(os.path.join('out_put', f'{cdiscName}.py'), "w", encoding="UTF-8") as f:
                f.write(output)

if __name__ == '__main__':
    f = "D:\\github\\pyodm\\src\\pyodm\\resources\odm\\schema\\ODM-protocol.xsd"
    g = GeneratePy(f)
    g.generate()
    g.write_spec(g.cdisc())
    g.write_cdisc(g.cdisc())




import pathlib

from lxml import etree

import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta
from pyodm.core.xml.writer.write_status import WriteStatus
from pyodm.exceptions import XmlWriterException


def cdisc(model):
    def __inner__(func):
        def __wrapper__(self, cdisc, element):
            if isinstance(cdisc, model):
                status = func(self, cdisc, element)
                if status == WriteStatus.IGNORE: return WriteStatus.IGNORE
            else:
                return WriteStatus.PASS
            return WriteStatus.OK

        return __wrapper__

    return __inner__


class CdiscXmlWriter:
    """
    ODM 对象导出为 XML 文件
    """
    def __init__(self, cdisc, out_put: pathlib.Path):
        self.cdisc = cdisc
        self.out_put = out_put

    def write(self):

        if not isinstance(self.cdisc, Meta.CdiscODMEntity):
            raise XmlWriterException("cdisc should instance of CdiscODMEntity")
        element = etree.Element(self.cdisc.get_name())

        for action in [self._entity, self._many_element, self._one_element]:
            status = action(cdisc=self.cdisc, element=element)
            if status != WriteStatus.PASS: break
        dump_tree = etree.ElementTree(element=element)
        dump_tree.write(self.out_put, pretty_print=True, encoding="utf-8")

    def cdisc_model(self, unknown) -> bool:
        return isinstance(unknown, Meta.CdiscODMEntity) or \
               isinstance(unknown, Model.Attribute) or \
               isinstance(unknown, Model.OneElement) or \
               isinstance(unknown, Model.ManyElements)

    @cdisc(model=Meta.CdiscODMEntity)
    def _entity(self, cdisc, element: etree._Element = None) -> WriteStatus:
        element = etree.Element(cdisc.get_name) if element is None else element
        if not cdisc.is_blank():
            element.text = cdisc.get_value()
        for name, cdisc_instance in vars(cdisc).items():
            if not self.cdisc_model(cdisc_instance): continue
            self._vars_cdisc(cdisc_instance, element=element)

    @cdisc(model=Model.Attribute)
    def _attribute(self, cdisc: Model.Attribute, element: etree._Element):
        if cdisc.no_use(): return WriteStatus.IGNORE
        element.attrib[cdisc.get_name()] = str(cdisc.get_value())

    @cdisc(model=Model.OneElement)
    def _one_element(self, cdisc: Model.OneElement, element: etree._Element) -> WriteStatus:
        if cdisc.no_use(): return WriteStatus.IGNORE
        for name, cdisc_instance in vars(cdisc).items():
            if not self.cdisc_model(cdisc_instance): continue
            if not cdisc.is_blank(): element.text = cdisc.get_value()
            self._vars_cdisc(cdisc_instance, element)

    @cdisc(model=Model.ManyElements)
    def _many_element(self, cdisc: Model.ManyElements, element: etree._Element) -> WriteStatus:
        if cdisc.no_use(): return WriteStatus.IGNORE
        for e in cdisc.array:
            self._entity(e, etree.SubElement(element, cdisc.get_name()))

    def _vars_cdisc(self, cdisc, element: etree._Element):
        for action in [self._attribute, self._many_element, self._one_element]:
            status = action(cdisc=cdisc, element=element)
            if status != WriteStatus.PASS: break
        if status == WriteStatus.PASS:
            self._entity(cdisc, etree.SubElement(element, cdisc.get_name()))



from pyodm.core.support.abstract_data_loader import AbstractDataLoader
from pyodm.core.support.abstract_data_reader import AbstractDataReader


class JsonDataReader(AbstractDataReader):
    def read_specified(self, specified_name: str, loader: AbstractDataLoader):
        json_data = loader.load()
        return self._parse(json_data, specified_name)

    def read(self, loader: AbstractDataLoader):
        ...

    def _new_node(self, name):
        new_obj = self.registry.get(name)()
        return new_obj

    def _parse(self, data: dict, name):
        cdisc_node = self._new_node(name)
        cdisc_node.set_name(name)
        for key, value in data.items():
            if isinstance(value, dict):
                # One Element
                one_element = self._parse(value, key)
                one_element.set_name(key)
                setattr(cdisc_node, key, one_element)
            elif isinstance(value, list):
                # Many Element
                sub_cdisc_node = getattr(cdisc_node, key)
                sub_cdisc_node.set_name(key)
                for e in value:
                    sub_cdisc_node << self._parse(e, key)
            else:
                # Attribute
                attr = getattr(cdisc_node, key)
                attr.set_name(key)
                attr.set_value(value)
        return cdisc_node



import pytest

from pyodm.core.xml.reader.configuration.cdisc_configuration_specification_reader import \
    CdiscConfigurationSpecificationReader
from pyodm.core.xml.reader.configuration.cdisc_configuration_xsd_reader import CdiscConfigurationXsdReader
from pyodm.factory.cdisc_registry import CdiscRegistry

from pyodm.utils.path_utils import PathUtils


@pytest.fixture
def xsd_files():
    base = PathUtils.folder("pyodm", __file__).joinpath("model", "v2", "resources", "schema")
    xsds = ["ODM-admindata.xsd", "ODM-clinicaldata.xsd", "ODM-foundation.xsd",
            "ODM-protocol.xsd", "ODM-referencedata.xsd", "ODM-study.xsd"]
    return [base.joinpath(xsd) for xsd in xsds]


@pytest.fixture
def specification_file():
    return PathUtils.folder("pyodm", __file__).joinpath("resources", "SpecificationV2.xml")


def test_with_xsd(xsd_files):
    """
    对应 https://wiki.cdisc.org/display/ODM2 检查 Element XPath(s)
    :param xsd_files:
    :type xsd_files:
    :return:
    :rtype:
    """
    registry = CdiscRegistry()
    CdiscConfigurationXsdReader().load_cdisc_definition(registry, xsd_files=xsd_files)
    # for e in registry.definition_tree.xpath("Signature", Direction.TO):
    #     print(e)

    for e in registry.definition_tree.xpath_range("ReferenceData", "Signature"):
        print(e)


def test_with_spec(specification_file):
    registry = CdiscRegistry()
    CdiscConfigurationSpecificationReader().load_cdisc_definition(registry, [specification_file])
    for e in registry.definition_tree.xpath("ODM"):
        print(e)
    registry.definition_tree.children_by_name("SubjectData")


import json
from pathlib import Path
import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta
from pyodm.core.xml.writer.write_status import WriteStatus
from pyodm.exceptions import XmlWriterException


def cdisc(model):
    def __inner__(func):
        def __wrapper__(self, cdisc, element):
            if isinstance(cdisc, model):
                status = func(self, cdisc, element)
                if status == WriteStatus.IGNORE: return WriteStatus.IGNORE
            else:
                return WriteStatus.PASS
            return WriteStatus.OK

        return __wrapper__

    return __inner__


class JSONWriter:
    def __init__(self, out_put: Path, data):
        self.out_put = out_put
        self._data = data

    def write(self):
        if not isinstance(self._data, Meta.CdiscODMEntity):
            raise XmlWriterException("data should instance of CdiscODMEntity")
        json_data = dict()

        for action in [self._entity, self._many_element, self._one_element]:
            status = action(self._data, json_data)
            if status != WriteStatus.PASS: break
        print(json_data)
        with open(self.out_put, 'w') as f:
            f.write(json.dumps(json_data, indent=4))
        # dump_tree.write(self.out_put, pretty_print=True, encoding="utf-8")

    def is_valid(self, unknown) -> bool:
        return isinstance(unknown, Meta.CdiscODMEntity) or \
               isinstance(unknown, Model.Attribute) or \
               isinstance(unknown, Model.OneElement) or \
               isinstance(unknown, Model.ManyElements)

    @cdisc(model=Meta.CdiscODMEntity)
    def _entity(self, entity, json_data: dict = None) -> WriteStatus:
        for name, entity_instance in vars(entity).items():
            if not self.is_valid(entity_instance): continue
            self._vars_entity(entity_instance, json_data=json_data)

    @cdisc(model=Model.Attribute)
    def _attribute(self, entity: Model.Attribute, json_data: dict = None):
        if entity.no_use(): return WriteStatus.IGNORE
        # json_data[entity.get_name()] = str(entity.get_value())
        json_data[entity.get_name()] = entity.get_value()

    @cdisc(model=Model.OneElement)
    def _one_element(self, entity: Model.OneElement, json_data: dict = None) -> WriteStatus:
        if entity.no_use(): return WriteStatus.IGNORE
        if not json_data.get(entity.get_name()): json_data[entity.get_name()] = dict
        for name, cdisc_instance in vars(cdisc).items():
            if not self.is_valid(cdisc_instance): continue
            # if not entity.is_blank(): element.text = cdisc.value
            self._vars_entity(cdisc_instance, json_data.get(entity.get_name()))

    @cdisc(model=Model.ManyElements)
    def _many_element(self, entity: Model.ManyElements, json_data: dict = None) -> WriteStatus:
        if entity.no_use(): return WriteStatus.IGNORE
        if not json_data.get(entity.get_name()): json_data[entity.get_name()] = list()
        for e in entity.array:
            object_data = dict()
            self._entity(e, object_data)
            json_data.get(entity.get_name()).append(object_data)

    def _vars_entity(self, entity, json_data: dict = None):
        for action in [self._attribute, self._many_element, self._one_element]:
            status = action(entity, json_data)
            if status != WriteStatus.PASS: break
        if status == WriteStatus.PASS:
            self._entity(entity, json_data.get(entity.get_name(), dict()))



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Address(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Address+Element
    """

    StreetName = Model.OneElement()

    HouseNumber = Model.OneElement()

    City = Model.OneElement()

    StateProv = Model.OneElement()

    Country = Model.OneElement()

    PostalCode = Model.OneElement()

    GeoPosition = Model.OneElement()

    OtherText = Model.OneElement()



import abc

from pyodm.core.support.abstract_data_loader import AbstractDataLoader
from pyodm.factory.cdisc_registry import CdiscRegistry


class AbstractDataReader(abc.ABC):

    """
    读去数据的抽象方法,定义抽象方法 read()
    """
    def __init__(self, registry: CdiscRegistry):
        self._registry = registry

    @property
    def registry(self): return self._registry

    @abc.abstractmethod
    def read(self, loader: AbstractDataLoader): ...

    def read_specified(self, specified_name: str, loader: AbstractDataLoader): ...



import pyodm.model.definition as Model






class Mixin:

    @classmethod
    def first(cls): return cls

    @classmethod
    def find(cls, **attributes): return cls

    @classmethod
    def index(cls, i): return cls

    @classmethod
    def each(cls, usage, **attributes): return cls

    @classmethod
    def select(cls, **attributes): return cls

    @classmethod
    def usage(cls, usage): return cls



class MetaDataVersion(Mixin):
    OID = Model.Attribute()
    Name = Model.Attribute()
    CommentOID = Model.Attribute()
    Description = Model.OneElement()
    Include = Model.OneElement()
    Standards = Model.OneElement()
    AnnotatedCRF = Model.OneElement()
    SupplementalDoc = Model.OneElement()
    ValueListDef = Model.ManyElements()
    WhereClauseDef = Model.ManyElements()
    Protocol = Model.OneElement()
    WorkflowDef = Model.ManyElements()
    StudyEventGroupDef = Model.ManyElements()
    StudyEventDef = Model.ManyElements()
    ItemGroupDef = Model.ManyElements()
    ItemDef = Model.ManyElements()
    CodeList = Model.ManyElements()
    ConditionDef = Model.ManyElements()
    MethodDef = Model.ManyElements()
    CommentDef = Model.ManyElements()
    Leaf = Model.ManyElements()


class ReferenceData(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ReferenceData
    """
    StudyOID = Model.Attribute()
    MetaDataVersionOID = Model.Attribute()
    ItemGroupData = Model.ManyElements()
    AuditRecord = Model.OneElement()
    Signature = Model.OneElement()
    Annotation = Model.ManyElements()


class AdminData(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/AdminData+Element
    """

    StudyOID = Model.Attribute()

    User = Model.ManyElements()

    Organization = Model.ManyElements()

    Location = Model.ManyElements()

    SignatureDef = Model.ManyElements()


class User(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/User+Element
    """

    OID = Model.Attribute()

    UserType = Model.Attribute()

    OrganizationOID = Model.Attribute()

    LocationOID = Model.Attribute()

    UserName = Model.OneElement()

    Prefix = Model.OneElement()

    Suffix = Model.OneElement()

    FullName = Model.OneElement()

    GivenName = Model.OneElement()

    FamilyName = Model.OneElement()

    Image = Model.OneElement()

    Address = Model.ManyElements()

    Telecom = Model.ManyElements()


class UserName(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/UserName+Element
    """


class Prefix(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Prefix+Element
    """


class Suffix(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Suffix+Element
    """


class FullName(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/FullName+Element
    """


class GivenName(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/GivenName+Element
    """


class FamilyName(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/FamilyName+Element
    """


class Image(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Image+Element
    """

    ImageFileName = Model.Attribute()

    href = Model.Attribute()

    MimeType = Model.Attribute()


class Organization(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Organization+Element
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Role = Model.Attribute()

    Type = Model.Attribute()

    LocationOID = Model.Attribute()

    PartOfOrganizationOID = Model.Attribute()

    Description = Model.OneElement()

    Address = Model.ManyElements()

    Telecom = Model.ManyElements()


class Location(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Location+Element
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Role = Model.Attribute()

    OrganizationOID = Model.Attribute()

    Description = Model.OneElement()

    MetaDataVersionRef = Model.ManyElements()

    Address = Model.ManyElements()

    Telecom = Model.ManyElements()

    Query = Model.ManyElements()


class Address(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Address+Element
    """

    StreetName = Model.OneElement()

    HouseNumber = Model.OneElement()

    City = Model.OneElement()

    StateProv = Model.OneElement()

    Country = Model.OneElement()

    PostalCode = Model.OneElement()

    GeoPosition = Model.OneElement()

    OtherText = Model.OneElement()


class Telecom(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Telecom+Element
    """

    TelecomType = Model.Attribute()

    Value = Model.Attribute()


class StreetName(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StreetName+Element
    """


class HouseNumber(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/HouseNumber+Element
    """


class City(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/City+Element
    """


class StateProv(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StateProv+Element
    """


class Country(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Country+Element
    """


class PostalCode(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/PostalCode+Element
    """


class GeoPosition(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/GeoPosition+Element
    """

    Longitude = Model.Attribute()

    Latitude = Model.Attribute()

    Altitude = Model.Attribute()


class OtherText(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/OtherText+Element
    """


class MetaDataVersionRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/MetaDataVersionRef+Element
    """

    StudyOID = Model.Attribute()

    MetaDataVersionOID = Model.Attribute()

    EffectiveDate = Model.Attribute()


class SignatureDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SignatureDef+Element
    """

    OID = Model.Attribute()

    Methodology = Model.Attribute()

    Meaning = Model.OneElement()

    LegalReason = Model.OneElement()


class Meaning(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Meaning+Element
    """


class LegalReason(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/LegalReason+Element
    """


class ClinicalData(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ClinicalData+Element
    """

    StudyOID = Model.Attribute()

    MetaDataVersionOID = Model.Attribute()

    SubjectData = Model.ManyElements()

    ItemGroupData = Model.ManyElements()

    Query = Model.ManyElements()

    AuditRecord = Model.OneElement()

    Signature = Model.OneElement()

    Annotation = Model.ManyElements()


class SubjectData(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SubjectData+Element
    """

    SubjectKey = Model.Attribute()

    TransactionType = Model.Attribute()

    InvestigatorRef = Model.OneElement()

    SiteRef = Model.OneElement()

    StudyEventData = Model.ManyElements()

    Query = Model.ManyElements()

    AuditRecord = Model.OneElement()

    Signature = Model.OneElement()

    Annotation = Model.ManyElements()


class SiteRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SiteRef+Element
    """

    LocationOID = Model.Attribute()


class InvestigatorRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/InvestigatorRef+Element
    """

    UserOID = Model.Attribute()


class StudyEventData(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEventData+Element
    """

    StudyEventOID = Model.Attribute()

    StudyEventRepeatKey = Model.Attribute()

    TransactionType = Model.Attribute()

    ItemGroupData = Model.ManyElements()

    Query = Model.ManyElements()

    AuditRecord = Model.OneElement()

    Signature = Model.OneElement()

    Annotation = Model.ManyElements()


class ItemGroupData(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ItemGroupData+Element
    """

    ItemGroupOID = Model.Attribute()

    ItemGroupRepeatKey = Model.Attribute()

    TransactionType = Model.Attribute()

    ItemGroupDataSeq = Model.Attribute()

    Query = Model.ManyElements()

    ItemGroupData = Model.ManyElements()

    ItemData = Model.ManyElements()

    AuditRecord = Model.OneElement()

    Signature = Model.OneElement()

    Annotation = Model.ManyElements()


class ItemData(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ItemData+Element
    """

    ItemOID = Model.Attribute()

    TransactionType = Model.Attribute()

    IsNull = Model.Attribute()

    Value = Model.ManyElements()

    Query = Model.ManyElements()

    AuditRecord = Model.OneElement()

    Signature = Model.OneElement()

    Annotation = Model.ManyElements()


class AuditRecord(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/AuditRecord+Element
    """

    EditPoint = Model.Attribute()

    UsedMethod = Model.Attribute()

    UserRef = Model.OneElement()

    LocationRef = Model.OneElement()

    DateTimeStamp = Model.OneElement()

    ReasonForChange = Model.OneElement()

    SourceID = Model.OneElement()


class UserRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/UserRef+Element
    """

    UserOID = Model.Attribute()


class LocationRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/LocationRef+Element
    """

    LocationOID = Model.Attribute()


class DateTimeStamp(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/DateTimeStamp+Element
    """


class ReasonForChange(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ReasonForChange+Element
    """


class SourceID(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SourceID+Element
    """


class Signature(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Signature+Element
    """

    ID = Model.Attribute()

    UserRef = Model.OneElement()

    LocationRef = Model.OneElement()

    SignatureRef = Model.OneElement()

    DateTimeStamp = Model.OneElement()


class SignatureRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SignatureRef+Element
    """

    SignatureOID = Model.Attribute()


class Alias(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Alias
    """

    Context = Model.Attribute()

    Name = Model.Attribute()


class Description(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Description
    """

    TranslatedText = Model.ManyElements()


class TranslatedText(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/TranslatedText
    """

    xml_lang = Model.Attribute()

    Type = Model.Attribute()

    xhtml_div = Model.ManyElements()


class Study(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Study
    """

    OID = Model.Attribute()

    StudyName = Model.Attribute()

    ProtocolName = Model.Attribute()

    VersionID = Model.Attribute()

    VersionName = Model.Attribute()

    Status = Model.Attribute()

    Description = Model.OneElement()

    MetaDataVersion = Model.ManyElements()


class MetaDataVersion(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/MetaDataVersion
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    CommentOID = Model.Attribute()

    Description = Model.OneElement()

    Include = Model.ManyElements()

    Standards = Model.ManyElements()

    AnnotatedCRF = Model.ManyElements()

    SupplementalDoc = Model.ManyElements()

    ValueListDef = Model.ManyElements()

    WhereClauseDef = Model.ManyElements()

    Protocol = Model.ManyElements()

    WorkflowDef = Model.ManyElements()

    StudyEventGroupDef = Model.ManyElements()

    StudyEventDef = Model.ManyElements()

    ItemGroupDef = Model.ManyElements()

    ItemDef = Model.ManyElements()

    CodeList = Model.ManyElements()

    ConditionDef = Model.ManyElements()

    MethodDef = Model.ManyElements()

    CommentDef = Model.ManyElements()

    Leaf = Model.ManyElements()


class DocumentRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/DocumentRef
    """

    LeafID = Model.Attribute()

    PDFPageRef = Model.ManyElements()


class PDFPageRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/PDFPageRef
    """


class Leaf(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Leaf
    """

    ID = Model.Attribute()

    xlink_href = Model.Attribute()

    Title = Model.ManyElements()


class Title(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Title
    """


class Include(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Include
    """

    StudyOID = Model.Attribute()

    MetaDataVersionOID = Model.Attribute()

    href = Model.Attribute()


class Standards(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Standards
    """

    Standard = Model.ManyElements()


class Standard(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Standard
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Type = Model.Attribute()

    PublishingSet = Model.Attribute()

    Version = Model.Attribute()

    Status = Model.Attribute()

    CommentOID = Model.Attribute()


class AnnotatedCRF(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/AnnotatedCRF
    """

    DocumentRef = Model.ManyElements()


class SupplementalDoc(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SupplementalDoc
    """

    DocumentRef = Model.ManyElements()


class ValueListDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ValueListDef
    """

    OID = Model.Attribute()

    Description = Model.ManyElements()

    ItemRef = Model.ManyElements()


class WhereClauseRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/WhereClauseRef
    """

    WhereClauseOID = Model.Attribute()


class WhereClauseDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/WhereClauseDef
    """

    OID = Model.Attribute()

    CommentOID = Model.Attribute()

    RangeCheck = Model.ManyElements()


class StudyEventGroupRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEventGroupRef
    """

    StudyEventGroupOID = Model.Attribute()

    OrderNumber = Model.Attribute()

    Mandatory = Model.Attribute()

    CollectionExceptionConditionOID = Model.Attribute()

    Description = Model.OneElement()


class StudyEventGroupDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEventGroupDef
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    ArmOID = Model.Attribute()

    EpochOID = Model.Attribute()

    CommentOID = Model.Attribute()

    Description = Model.OneElement()

    WorkflowRef = Model.OneElement()

    Coding = Model.ManyElements()

    StudyEventGroupRef = Model.OneElement()

    StudyEventRef = Model.OneElement()


class StudyEventRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEventRef
    """

    StudyEventOID = Model.Attribute()

    OrderNumber = Model.Attribute()

    Mandatory = Model.Attribute()

    CollectionExceptionConditionOID = Model.Attribute()


class StudyEventDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEventDef
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Repeating = Model.Attribute()

    Type = Model.Attribute()

    Category = Model.Attribute()

    CommentOID = Model.Attribute()

    Description = Model.OneElement()

    ItemGroupRef = Model.ManyElements()

    WorkflowRef = Model.OneElement()

    Coding = Model.ManyElements()

    Alias = Model.ManyElements()


class ItemGroupRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ItemGroupRef
    """

    ItemGroupOID = Model.Attribute()

    MethodOID = Model.Attribute()

    OrderNumber = Model.Attribute()

    Mandatory = Model.Attribute()

    CollectionExceptionConditionOID = Model.Attribute()


class ItemGroupDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ItemGroupDef
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Repeating = Model.Attribute()

    RepeatingLimit = Model.Attribute()

    IsReferenceData = Model.Attribute()

    Structure = Model.Attribute()

    ArchiveLocationID = Model.Attribute()

    DatasetName = Model.Attribute()

    Domain = Model.Attribute()

    Type = Model.Attribute()

    Purpose = Model.Attribute()

    StandardOID = Model.Attribute()

    IsNonStandard = Model.Attribute()

    HasNoData = Model.Attribute()

    CommentOID = Model.Attribute()

    Description = Model.OneElement()

    Class = Model.ManyElements()

    Coding = Model.ManyElements()

    WorkflowRef = Model.ManyElements()

    Origin = Model.ManyElements()

    Alias = Model.ManyElements()

    Leaf = Model.ManyElements()

    ItemGroupRef = Model.OneElement()

    ItemRef = Model.OneElement()


class Class(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Class
    """

    Name = Model.Attribute()

    SubClass = Model.ManyElements()


class SubClass(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SubClass
    """


class ItemRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ItemRef
    """

    ItemOID = Model.Attribute()

    KeySequence = Model.Attribute()

    IsNonStandard = Model.Attribute()

    HasNoData = Model.Attribute()

    MethodOID = Model.Attribute()

    UnitsItemOID = Model.Attribute()

    Repeat = Model.Attribute()

    Other = Model.Attribute()

    Role = Model.Attribute()

    RoleCodeListOID = Model.Attribute()

    Core = Model.Attribute()

    PreSpecifiedValue = Model.Attribute()

    OrderNumber = Model.Attribute()

    Mandatory = Model.Attribute()

    CollectionExceptionConditionOID = Model.Attribute()

    Origin = Model.ManyElements()

    WhereClauseRef = Model.ManyElements()


class Origin(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Origin
    """

    Type = Model.Attribute()

    Source = Model.Attribute()

    Description = Model.OneElement()

    SourceItems = Model.OneElement()

    Coding = Model.ManyElements()

    DocumentRef = Model.ManyElements()


class SourceItems(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SourceItems
    """

    SourceItem = Model.ManyElements()

    Coding = Model.ManyElements()


class SourceItem(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SourceItem
    """

    ItemOID = Model.Attribute()

    ItemGroupOID = Model.Attribute()

    MetaDataVersionOID = Model.Attribute()

    StudyOID = Model.Attribute()

    leafID = Model.Attribute()

    Name = Model.Attribute()

    Resource = Model.ManyElements()

    Coding = Model.ManyElements()


class Resource(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Resource
    """

    Type = Model.Attribute()

    Name = Model.Attribute()

    Attribute = Model.Attribute()

    Label = Model.Attribute()

    Selection = Model.ManyElements()


class Selection(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Selection
    """

    Path = Model.Attribute()


class ItemDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ItemDef
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    DataType = Model.Attribute()

    Length = Model.Attribute()

    DisplayFormat = Model.Attribute()

    VariableSet = Model.Attribute()

    CommentOID = Model.Attribute()

    Description = Model.OneElement()

    Definition = Model.OneElement()

    Question = Model.OneElement()

    Prompt = Model.OneElement()

    CRFCompletionInstructions = Model.OneElement()

    ImplementationNotes = Model.OneElement()

    CDISCNotes = Model.OneElement()

    RangeCheck = Model.ManyElements()

    CodeListRef = Model.ManyElements()

    ValueListRef = Model.ManyElements()

    Coding = Model.ManyElements()

    Alias = Model.ManyElements()


class Question(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Question
    """

    TranslatedText = Model.ManyElements()


class Definition(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Definition
    """

    TranslatedText = Model.ManyElements()


class Prompt(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Prompt
    """

    TranslatedText = Model.ManyElements()


class CRFCompletionInstructions(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/CRFCompletionInstructions
    """

    TranslatedText = Model.ManyElements()


class ImplementationNotes(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ImplementationNotes
    """

    TranslatedText = Model.ManyElements()


class CDISCNotes(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/CDISCNotes
    """

    TranslatedText = Model.ManyElements()


class RangeCheck(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/RangeCheck
    """

    Comparator = Model.Attribute()

    SoftHard = Model.Attribute()

    ItemOID = Model.Attribute()

    ErrorMessage = Model.ManyElements()


class CheckValue(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/CheckValue
    """


class ErrorMessage(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ErrorMessage
    """

    TranslatedText = Model.ManyElements()


class CodeListRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/CodeListRef
    """

    CodeListOID = Model.Attribute()


class ValueListRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ValueListRef
    """

    ValueListOID = Model.Attribute()


class CodeList(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/CodeList
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    DataType = Model.Attribute()

    CommentOID = Model.Attribute()

    StandardOID = Model.Attribute()

    IsNonStandard = Model.Attribute()

    Description = Model.OneElement()

    CodeListItem = Model.ManyElements()

    Coding = Model.ManyElements()

    Alias = Model.ManyElements()


class CodeListItem(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/CodeListItem
    """

    CodedValue = Model.Attribute()

    Rank = Model.Attribute()

    Other = Model.Attribute()

    OrderNumber = Model.Attribute()

    ExtendedValue = Model.Attribute()

    CommentOID = Model.Attribute()

    Description = Model.OneElement()

    Decode = Model.OneElement()

    Coding = Model.ManyElements()

    Alias = Model.ManyElements()


class Decode(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Decode
    """

    TranslatedText = Model.ManyElements()


class MethodDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/MethodDef
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Type = Model.Attribute()

    CommentOID = Model.Attribute()

    Description = Model.OneElement()

    MethodSignature = Model.ManyElements()

    FormalExpression = Model.ManyElements()

    Alias = Model.ManyElements()

    DocumentRef = Model.ManyElements()


class MethodSignature(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/MethodSignature
    """

    Parameter = Model.ManyElements()

    ReturnValue = Model.ManyElements()


class Parameter(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Parameter
    """

    Name = Model.Attribute()

    DataType = Model.Attribute()

    Definition = Model.Attribute()

    OrderNumber = Model.Attribute()


class ReturnValue(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ReturnValue
    """

    Name = Model.Attribute()

    DataType = Model.Attribute()

    Definition = Model.Attribute()

    OrderNumber = Model.Attribute()


class ConditionDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ConditionDef
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    CommentOID = Model.Attribute()

    Description = Model.OneElement()

    MethodSignature = Model.ManyElements()

    FormalExpression = Model.ManyElements()

    Alias = Model.ManyElements()


class FormalExpression(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/FormalExpression
    """

    Context = Model.Attribute()


class Code(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Code
    """


class ExternalCodeLib(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ExternalCodeLib
    """

    Library = Model.Attribute()

    Method = Model.Attribute()

    Version = Model.Attribute()

    ref = Model.Attribute()

    href = Model.Attribute()


class CommentDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/CommentDef
    """

    OID = Model.Attribute()

    Description = Model.OneElement()

    DocumentRef = Model.ManyElements()


class ODM(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ODM
    """

    FileType = Model.Attribute()

    Granularity = Model.Attribute()

    Context = Model.Attribute()

    FileOID = Model.Attribute()

    CreationDateTime = Model.Attribute()

    PriorFileOID = Model.Attribute()

    AsOfDateTime = Model.Attribute()

    ODMVersion = Model.Attribute()

    Originator = Model.Attribute()

    SourceSystem = Model.Attribute()

    SourceSystemVersion = Model.Attribute()

    Description = Model.ManyElements()

    Study = Model.ManyElements()

    AdminData = Model.ManyElements()

    ReferenceData = Model.ManyElements()

    ClinicalData = Model.ManyElements()

    Association = Model.ManyElements()


class Association(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Association
    """

    StudyOID = Model.Attribute()

    MetaDataVersionOID = Model.Attribute()

    # 手动处理了，因为标准定义了2个相同的子元素
    KeySet = Model.ManyElements()

    Annotation = Model.OneElement()


class KeySet(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/KeySet
    """

    StudyOID = Model.Attribute()

    SubjectKey = Model.Attribute()

    MetaDataVersionOID = Model.Attribute()

    StudyEventOID = Model.Attribute()

    StudyEventRepeatKey = Model.Attribute()

    ItemGroupOID = Model.Attribute()

    ItemGroupRepeatKey = Model.Attribute()

    ItemOID = Model.Attribute()


class Annotation(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Annotation
    """

    SeqNum = Model.Attribute()

    TransactionType = Model.Attribute()

    ID = Model.Attribute()

    Comment = Model.OneElement()

    Coding = Model.ManyElements()

    Flag = Model.ManyElements()


class Comment(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Comment
    """

    SponsorOrSite = Model.Attribute()

    TranslatedText = Model.ManyElements()


class Flag(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Flag
    """

    FlagValue = Model.OneElement()

    FlagType = Model.OneElement()


class FlagValue(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/FlagValue
    """

    CodeListOID = Model.Attribute()


class FlagType(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/FlagType
    """

    CodeListOID = Model.Attribute()


class Coding(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Coding
    """

    Code = Model.Attribute()

    System = Model.Attribute()

    SystemName = Model.Attribute()

    SystemVersion = Model.Attribute()

    Label = Model.Attribute()

    href = Model.Attribute()

    ref = Model.Attribute()

    CommentOID = Model.Attribute()


class Query(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Query
    """

    OID = Model.Attribute()

    Source = Model.Attribute()

    Target = Model.Attribute()

    Type = Model.Attribute()

    State = Model.Attribute()

    LastUpdateDatetime = Model.Attribute()

    Name = Model.Attribute()

    Value = Model.OneElement()

    AuditRecord = Model.ManyElements()


class Value(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Value
    """

    SeqNum = Model.Attribute()


class Protocol(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Protocol
    """

    Description = Model.OneElement()

    StudySummary = Model.OneElement()

    StudyStructure = Model.OneElement()

    TrialPhase = Model.OneElement()

    StudyTimings = Model.OneElement()

    StudyIndications = Model.OneElement()

    StudyInterventions = Model.OneElement()

    StudyObjectives = Model.OneElement()

    StudyEndPoints = Model.OneElement()

    StudyTargetPopulation = Model.OneElement()

    StudyEstimands = Model.OneElement()

    InclusionExclusionCriteria = Model.OneElement()

    StudyEventGroupRef = Model.ManyElements()

    WorkflowRef = Model.OneElement()

    Alias = Model.ManyElements()


class StudyStructure(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyStructure
    """

    Description = Model.OneElement()

    Arm = Model.ManyElements()

    Epoch = Model.ManyElements()

    WorkflowRef = Model.OneElement()


class TrialPhase(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/TrialPhase
    """

    Value = Model.Attribute()

    Description = Model.OneElement()


class StudyIndications(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyIndications
    """

    StudyIndication = Model.ManyElements()


class StudyIndication(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyIndication
    """

    OID = Model.Attribute()

    Description = Model.OneElement()

    Coding = Model.ManyElements()


class StudyInterventions(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyInterventions
    """

    StudyIntervention = Model.ManyElements()


class StudyIntervention(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyIntervention
    """

    OID = Model.Attribute()

    Description = Model.OneElement()

    Coding = Model.ManyElements()


class StudyObjectives(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyObjectives
    """

    StudyObjective = Model.ManyElements()


class StudyObjective(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyObjective
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Level = Model.Attribute()

    Description = Model.OneElement()

    StudyEndPointRef = Model.ManyElements()


class StudyEndPointRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEndPointRef
    """

    StudyEndPointOID = Model.Attribute()

    OrderNumber = Model.Attribute()


class StudyEndPoints(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEndPoints
    """

    StudyEndPoint = Model.ManyElements()


class StudyEndPoint(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEndPoint
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Type = Model.Attribute()

    Level = Model.Attribute()

    Description = Model.OneElement()

    FormalExpression = Model.ManyElements()


class StudyTargetPopulation(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyTargetPopulation
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Description = Model.OneElement()

    Coding = Model.ManyElements()

    FormalExpression = Model.ManyElements()


class StudyEstimands(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEstimands
    """

    StudyEstimand = Model.ManyElements()


class StudyEstimand(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyEstimand
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Level = Model.Attribute()

    Description = Model.OneElement()

    StudyTargetPopulationRef = Model.OneElement()

    StudyInterventionRef = Model.OneElement()

    StudyEndPointRef = Model.OneElement()

    IntercurrentEvent = Model.ManyElements()

    SummaryMeasure = Model.OneElement()


class InclusionExclusionCriteria(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/InclusionExclusionCriteria
    """

    InclusionCriteria = Model.OneElement()

    ExclusionCriteria = Model.OneElement()


class InclusionCriteria(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/InclusionCriteria
    """

    Criterion = Model.ManyElements()


class ExclusionCriteria(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ExclusionCriteria
    """

    Criterion = Model.ManyElements()


class StudyTargetPopulationRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyTargetPopulationRef
    """

    StudyTargetPopulationOID = Model.Attribute()


class StudyInterventionRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyInterventionRef
    """

    StudyInterventionOID = Model.Attribute()


class IntercurrentEvent(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/IntercurrentEvent
    """

    Description = Model.OneElement()


class SummaryMeasure(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/SummaryMeasure
    """

    Description = Model.OneElement()


class Arm(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Arm
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Description = Model.OneElement()

    WorkflowRef = Model.OneElement()


class Epoch(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Epoch
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    SequenceNumber = Model.Attribute()

    Description = Model.OneElement()


class WorkflowRef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/WorkflowRef
    """

    WorkflowOID = Model.Attribute()


class StudySummary(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudySummary
    """

    StudyParameter = Model.ManyElements()


class StudyParameter(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyParameter
    """

    OID = Model.Attribute()

    Term = Model.Attribute()

    ShortName = Model.Attribute()

    ParameterValue = Model.ManyElements()

    Coding = Model.ManyElements()


class ParameterValue(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/ParameterValue
    """

    Value = Model.Attribute()

    Coding = Model.ManyElements()


class StudyTimings(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyTimings
    """

    StudyTiming = Model.ManyElements()


class StudyTiming(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/StudyTiming
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    AbsoluteTimingConstraint = Model.ManyElements()

    RelativeTimingConstraint = Model.ManyElements()

    TransitionTimingConstraint = Model.ManyElements()

    DurationTimingConstraint = Model.ManyElements()


class TransitionTimingConstraint(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/TransitionTimingConstraint
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    TransitionOID = Model.Attribute()

    MethodOID = Model.Attribute()

    Type = Model.Attribute()

    TimepointTarget = Model.Attribute()

    TimepointPreWindow = Model.Attribute()

    TimepointPostWindow = Model.Attribute()

    Description = Model.OneElement()


class AbsoluteTimingConstraint(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/AbsoluteTimingConstraint
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    StudyEventGroupOID = Model.Attribute()

    StudyEventOID = Model.Attribute()

    TimepointTarget = Model.Attribute()

    TimepointPreWindow = Model.Attribute()

    TimepointPostWindow = Model.Attribute()

    Description = Model.OneElement()


class RelativeTimingConstraint(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/RelativeTimingConstraint
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    PredecessorOID = Model.Attribute()

    SuccessorOID = Model.Attribute()

    Type = Model.Attribute()

    TimepointRelativeTarget = Model.Attribute()

    TimepointPreWindow = Model.Attribute()

    TimepointPostWindow = Model.Attribute()

    Description = Model.OneElement()


class DurationTimingConstraint(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/DurationTimingConstraint
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    StructuralElementOID = Model.Attribute()

    DurationTarget = Model.Attribute()

    DurationPreWindow = Model.Attribute()

    DurationPostWindow = Model.Attribute()

    Description = Model.OneElement()


class WorkflowDef(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/WorkflowDef
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Description = Model.OneElement()

    WorkflowStart = Model.OneElement()

    WorkflowEnd = Model.ManyElements()

    Transition = Model.OneElement()

    Branching = Model.OneElement()


class WorkflowStart(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/WorkflowStart
    """

    StartOID = Model.Attribute()


class Transition(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Transition
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    SourceOID = Model.Attribute()

    TargetOID = Model.Attribute()

    StartConditionOID = Model.Attribute()

    EndConditionOID = Model.Attribute()


class Branching(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Branching
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    Type = Model.Attribute()

    TargetTransition = Model.ManyElements()

    DefaultTransition = Model.ManyElements()


class TargetTransition(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/TargetTransition
    """

    TargetTransitionOID = Model.Attribute()

    ConditionOID = Model.Attribute()


class DefaultTransition(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/DefaultTransition
    """

    TargetTransitionOID = Model.Attribute()


class WorkflowEnd(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/WorkflowEnd
    """

    EndOID = Model.Attribute()


class Criterion(Mixin):
    """
    https://wiki.cdisc.org/display/ODM2/Criterion
    """

    OID = Model.Attribute()

    Name = Model.Attribute()

    ConditionOID = Model.Attribute()

    Description = Model.OneElement()

    Coding = Model.ManyElements()


class Specification:
    """
    定义 CDISC ODM V2 的节点，该文件属于结构的声明文件，真实实现在 cdisc 中
    """
    Association: Association
    KeySet: KeySet
    Annotation: Annotation
    Comment: Comment
    Flag: Flag
    FlagValue: FlagValue
    FlagType: FlagType
    Coding: Coding
    Query: Query
    Value: Value
    User: User
    UserName: UserName
    Prefix: Prefix
    Suffix: Suffix
    FullName: FullName
    GivenName: GivenName
    FamilyName: FamilyName
    Image: Image
    Organization: Organization
    Location: Location
    Address: Address
    Telecom: Telecom
    StreetName: StreetName
    HouseNumber: HouseNumber
    City: City
    StateProv: StateProv
    Country: Country
    PostalCode: PostalCode
    GeoPosition: GeoPosition
    OtherText: OtherText
    MetaDataVersionRef: MetaDataVersionRef
    SignatureDef: SignatureDef
    Meaning: Meaning
    LegalReason: LegalReason
    ClinicalData: ClinicalData
    SubjectData: SubjectData
    SiteRef: SiteRef
    InvestigatorRef: InvestigatorRef
    StudyEventData: StudyEventData
    ItemGroupData: ItemGroupData
    ItemData: ItemData
    AuditRecord: AuditRecord
    UserRef: UserRef
    LocationRef: LocationRef
    DateTimeStamp: DateTimeStamp
    ReasonForChange: ReasonForChange
    SourceID: SourceID
    Signature: Signature
    SignatureRef: SignatureRef
    FormalExpression: FormalExpression
    Code: Code
    ExternalCodeLib: ExternalCodeLib
    CommentDef: CommentDef
    StudyStructure: StudyStructure
    TrialPhase: TrialPhase
    StudyIndications: StudyIndications
    StudyIndication: StudyIndication
    StudyInterventions: StudyInterventions
    StudyIntervention: StudyIntervention
    StudyObjectives: StudyObjectives
    StudyObjective: StudyObjective
    StudyEndPointRef: StudyEndPointRef
    StudyEndPoints: StudyEndPoints
    StudyEndPoint: StudyEndPoint
    StudyTargetPopulation: StudyTargetPopulation
    StudyEstimands: StudyEstimands
    StudyEstimand: StudyEstimand
    InclusionExclusionCriteria: InclusionExclusionCriteria
    InclusionCriteria: InclusionCriteria
    ExclusionCriteria: ExclusionCriteria
    StudyTargetPopulationRef: StudyTargetPopulationRef
    StudyInterventionRef: StudyInterventionRef
    IntercurrentEvent: IntercurrentEvent
    SummaryMeasure: SummaryMeasure
    Arm: Arm
    Epoch: Epoch
    WorkflowRef: WorkflowRef
    StudySummary: StudySummary
    StudyParameter: StudyParameter
    ParameterValue: ParameterValue
    StudyTimings: StudyTimings
    StudyTiming: StudyTiming
    TransitionTimingConstraint: TransitionTimingConstraint
    AbsoluteTimingConstraint: AbsoluteTimingConstraint
    RelativeTimingConstraint: RelativeTimingConstraint
    DurationTimingConstraint: DurationTimingConstraint
    WorkflowDef: WorkflowDef
    WorkflowStart: WorkflowStart
    Transition: Transition
    Branching: Branching
    TargetTransition: TargetTransition
    DefaultTransition: DefaultTransition
    WorkflowEnd: WorkflowEnd
    Criterion: Criterion

    def first(self): return self

    def find(self): return self

    def index(self, i): return self



import abc

from pyodm.core.support.abstract_data_loader import AbstractDataLoader
from pyodm.core.support.abstract_data_reader import AbstractDataReader

from pyodm.factory.cdisc_registry import CdiscRegistry


class AbstractDataObjectFactory(abc.ABC, CdiscRegistry):

    def __init__(self):
        super().__init__()
        self._data_object = None
        self._data_loader = None

    @abc.abstractmethod
    def clazz_reader(self):
        """
        指定 生成 ODM Class 的方式
        :return:
        :rtype:
        """
        ...

    @abc.abstractmethod
    def data_reader(self) -> AbstractDataReader:
        """
        指定 解析 ODM Data 的方式
        :return:
        :rtype:
        """
        ...

    @property
    def data_loader(self):
        """
        读取 ODM 数据
        :return:
        :rtype:
        """
        return self._data_loader

    @data_loader.setter
    def data_loader(self, data_loader: AbstractDataLoader):
        self._data_loader = data_loader

    def _data_process(self, specified: str = None, refresh: bool = False):
        """
        解析 ODM Data 模板方法
        :param specified: 指定 要读取的 ODM 对象
        :type: str
        :param refresh: 是否重新读取 True Y, False N
        :type refresh: bool
        :return:
        :rtype:
        """
        __new = True if self._data_object is None else refresh
        if __new:
            self.clazz_reader()
            if specified is None:
                self._data_object = self.data_reader().read(self.data_loader)
            else:
                self._data_object = self.data_reader().read_specified(specified, self.data_loader)
        return self._data_object

    def factory(self, specified: str = None, refresh: bool = False):
        return self._data_process(specified, refresh)



from pyodm.exceptions import ParserException
from pyodm.factory.xpath.direction import Direction
from pyodm.factory.xpath.xpath import Xpath
from pyodm.factory.xpath.xpath_node import XpathNode


class XpathTree:
    """
    特殊的树形结构,非通用数据结构,仅适用 XPATH 功能
    目标 : 解析并保存 Class 的树形结构
    """
    def __init__(self):
        self._nodes: list[XpathNode] = []

    def count(self):
        return len(self._nodes)

    def add(self, parent, children):
        self._add_parent(parent)
        self._add_child(parent, children)

    def find(self, node: XpathNode) -> int:
        try:
            return self._nodes.index(node)
        except ValueError as e:
            return -1

    def _add_parent(self, parent):
        i = self.find(parent)
        if i == -1:
            self._nodes.append(parent)

    def _add_child(self, parent, children):
        # 获取父节点索引
        i = self.find(parent)
        for child in children:
            j = self.find(child)
            if j == -1:
                # 如果子节点不存在
                self._nodes.append(child)
                child.parent.append(i)
                self._nodes[i].children.append(self.count() - 1)
            else:
                # 如果子节点已存在
                self._nodes[j].parent.append(i)
                self._nodes[i].children.append(j)

    def children_by_name(self, node_name: str) -> list[XpathNode]:
        for node in self._nodes:
            if node.name == node_name:
                return [self._nodes[i] for i in node.children]
        return []

    def children_by_node(self, node: XpathNode) -> list[XpathNode]:
        return [self._nodes[i] for i in node.children]

    def roots(self) -> list:
        results = []
        for i, node in enumerate(self._nodes):
            if node.isRoot(): results.append(i)
        return results

    def _sub_paths(self, node: int) -> list:
        path = []
        for child in self._nodes[node].children:
            if child == node: continue  # 如果子节点含有父亲节点
            sub_path = [node]
            if self._nodes[child].children:
                for branch in self._sub_paths(child):
                    new_branch = sub_path.copy()
                    new_branch.extend(branch)
                    if new_branch not in path:
                        path.append(new_branch)
            else:
                sub_path.append(child)
                if sub_path not in path:
                    path.append(sub_path)
        return path

    def _all_path_index(self):
        path = []
        for root in self.roots():
            if self._nodes[root].children:
                new_branch = self._sub_paths(root)
                if new_branch not in path:
                    path.extend(self._sub_paths(root))
        return path

    def _all_path_name(self):
        return [[self._nodes[i].name for i in indexes] for indexes in self._all_path_index()]

    def xpath(self, node_name=None, direction: Direction = Direction.FROM):
        paths = self._all_path_name()
        if node_name is not None:
            if direction == Direction.FROM:
                paths = Xpath(paths).start(name=node_name).paths
            elif direction == Direction.TO:
                paths = Xpath(paths).end(name=node_name).paths
            elif direction == Direction.INCLUDE:
                paths = Xpath(paths).contain(name=node_name).paths
            else:
                raise ParserException("direction is wrong")
        return list(set(["/".join(path) for path in paths]))

    def xpath_range(self, start: str, end: str):
        paths = self._all_path_name()
        paths = Xpath(paths).start(start).end(end).paths
        return list(set(["/".join(path) for path in paths]))



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class GeoPosition(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/GeoPosition+Element
    """
    
    Longitude = Model.Attribute()
    
    Latitude = Model.Attribute()
    
    Altitude = Model.Attribute()
    
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class DurationTimingConstraint(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/DurationTimingConstraint
    """
    
    OID = Model.Attribute()
    
    Name = Model.Attribute()
    
    StructuralElementOID = Model.Attribute()
    
    DurationTarget = Model.Attribute()
    
    DurationPreWindow = Model.Attribute()
    
    DurationPostWindow = Model.Attribute()
    
    
    Description = Model.OneElement()
    



from pyodm.core.data_loader.forest_data_loader import ForestDataLoader
from pyodm.core.support.abstract_data_reader import AbstractDataReader
from pyodm.factory.cdisc_registry import CdiscRegistry
from pyodm.model.definition import ManyElements, OneElement, Attribute
from pyodm.model.meta.cdisc_odm_entity import CdiscODMEntity
from pyodm.utils.forest import Forest


class ForestDataReader(AbstractDataReader):
    """
    读取Forest类型数据，把 Forest 类型转化为 ODM 数据
    """

    def __init__(self, registry: CdiscRegistry):
        super().__init__(registry)

    def read(self, data_loader: ForestDataLoader):
        forest_data = data_loader.load()
        result = self.read_data(forest_data)
        if len(result) == 1: return result[0]
        return result

    def read_data(self, forest_data: Forest):
        # self.otree = otree
        result = []
        for node in forest_data.get_level_nodes(0):
            for node_name, node_value in node.values.items():
                result.append(self._scan(node, node_name, node_value))
        return result

    def _sub_cdisc_node(self, cdisc_node, sub_cdisc_name):
        return getattr(cdisc_node, sub_cdisc_name)

    def _new_cdisc_node(self, cdisc_name):
        new_obj = self.registry.get(cdisc_name)()
        new_obj.set_name(cdisc_name)
        return new_obj

    def _scan(self, node, cdisc_name, node_value):
        cdisc = self._new_cdisc_node(cdisc_name)
        for key, value in node_value.items():
            if key == "Text":
                cdisc.set_value(value)
            else:
                attribute = getattr(cdisc, key)
                if attribute is None: raise Exception(f'{cdisc.get_name} have not attribute {key}')
                if not isinstance(attribute, Attribute): raise Exception(f'{cdisc.get_name} {key} is not Attribute')
                attribute.set_name(key)
                attribute.set_value(value)

        for sub_node in node.children():
            for sub_node_name, sub_node_value in sub_node.values.items():
                element = getattr(cdisc, sub_node_name, None)
                # if element is None: raise Exception(f'{cdisc.name} have not element {sub_node_name}')
                if element is None: continue
                if not isinstance(element, OneElement) and not isinstance(element, ManyElements) \
                        and not isinstance(element, CdiscODMEntity):
                    raise Exception(f'{cdisc.get_name()} {sub_node_name} is not element')
                if isinstance(element, OneElement):
                    setattr(cdisc, sub_node_name, self._scan(sub_node, sub_node_name, sub_node_value))
                elif isinstance(element, ManyElements):
                    element.set_name(sub_node_name)
                    element << self._scan(sub_node, sub_node_name, sub_node_value)
        return cdisc

    def _parse(self, cdisc, node_value: dict, otree_index=0):
        for attr_name, attr_obj in vars(cdisc).items():
            if node_value.get(attr_name) is not None:
                attr_obj.set_name(attr_name)
                attr_obj.set_value(node_value.get(attr_name))
        if node_value.get("Text") is not None:
            cdisc.set_value(node_value.get("Text"))
        self._scan(cdisc.get_name(), otree_index)
        return cdisc

    def _set_text(self, node_name, attr_name, info_dict, cdisc_node):
        if node_name == attr_name:
            if info_dict.get("Text") is not None:
                cdisc_node.set_value(info_dict.get("Text"))
                return True
        return False

    def _set_attribute(self, node_name, attr_name, node, cdisc_node):
        if node_name == attr_name:
            for name, attr_obj in vars(cdisc_node).items():
                if node.get(name) is not None:
                    attr_obj.set_name(name)
                    attr_obj.set_value(node.values.get(name))

    def _many_element(self, node_name, attr_name, sub_cdisc_node, otree_index):
        if isinstance(sub_cdisc_node, ManyElements):
            sub_cdisc_node << self._parse(self.registry.get(sub_cdisc_node.get_name()), otree_index)

    def _one_element(self, node_name, attr_name, sub_cdisc_node, otree_index):
        if isinstance(sub_cdisc_node, OneElement):
            self._parse(self.registry.get(sub_cdisc_node.get_name()), otree_index)




class Finder:
    def __init__(self):
        self._element_name = None
        self._attribute = dict()
        self._text = None

    def element(self, name: str):
        self._element_name = name
        return self

    def attribute(self, **kwargs):
        self._attribute.update(kwargs)
        return self

    def text(self, value):
        self._text = value
        return self

    def _attribute_match(self, element):
        for key, value in self._attribute.items():
            if element.__dict__.get(key) is None: return False
            if element.__dict__.get(key).get_value() != value: return False
        return True

    def _text_match(self, element):
        if self._text is None: return True
        else: return element.get_value() == self._text

    def _name_match(self, element):
        return element.get_name() == self._element_name

    def Lambda(self):
        return lambda x: self._name_match(x) and self._attribute_match(x) and self._text_match(x)



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class ItemData(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/ItemData+Element
    """
    
    ItemOID = Model.Attribute()
    
    TransactionType = Model.Attribute()

    IsNull = Model.Attribute()

    Value = Model.ManyElements()
    
    Query = Model.ManyElements()

    AuditRecord = Model.ManyElements()

    



from pyodm.core.support.abstract_path_loader import AbstractPathDataLoader
import json


class JsonPathLoader(AbstractPathDataLoader):
    def load(self):
        with open(self.file, 'r') as f:
            return json.load(f)



import abc

from pyodm.factory.abstract_cdisc_factory import AbstractCdiscFactory


class AbstractHierarchyFactory(AbstractCdiscFactory, abc.ABC):
    """
    针对 数据源类似 XML 读出数据，
    因为这类型的数据之间包含了层级关系
    """
    def __init__(self, configuration_files: list = None):
        """
        :param configuration_files: 默认是使用 XSD配置文件
        :type configuration_files:
        """
        super().__init__()
        self._configuration_files = configuration_files if configuration_files else self.default_description_files()



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class SourceItem(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/SourceItem
    """
    
    ItemOID = Model.Attribute()
    
    ItemGroupOID = Model.Attribute()
    
    MetaDataVersionOID = Model.Attribute()
    
    StudyOID = Model.Attribute()
    
    leafID = Model.Attribute()
    
    Name = Model.Attribute()
    
    
    Resource = Model.ManyElements()
    
    Coding = Model.ManyElements()
    



import abc

from pyodm.factory.abstract_data_object_factory import AbstractDataObjectFactory

from pyodm.model import odm_xsd_description


class AbstractCdiscFactory(AbstractDataObjectFactory, abc.ABC):
    """
    生成 ODM 的抽象工厂,定义类生成ODM实例的模板方法

    """

    def default_description_files(self) -> list:
        """
        默认的 XSD 文件路径 , 如果没有指定
        clazz_reader,默认采取的是XSD 的方式解析 ODM Class
        :return:
        :rtype:
        """
        return odm_xsd_description()

    def odm(self, refresh: bool = False):
        """
        返回 ODM 对象
        :param refresh:
        :type refresh:
        :return:
        :rtype:
        """
        return self.factory(refresh=refresh)



from pyodm.core.data_loader.forest_data_loader import ForestDataLoader
from pyodm.core.forest.reader.forest_data_reader import ForestDataReader

from pyodm.core.support.abstract_data_source import AbstractDataSource

from pyodm.factory.abstract_assemble_factory import AbstractAssembleFactory


class CdiscListableFactory(AbstractAssembleFactory):
    """
    针对 数据源类似从数据库读出的数据，
    因为这类型的数据之间并没有层级关系，需要把平行数据转化为树形数据的情况
    """

    def __init__(self, source: AbstractDataSource, configuration_files: list = None):
        """

        :param source:
        :type source:
        :param configuration_files: 默认是使用 XSD配置文件
        :type configuration_files:
        """
        super().__init__()
        self._configuration_files = configuration_files if configuration_files else self.default_description_files()
        source.registry = self
        self.data_loader = ForestDataLoader(source)

    def data_reader(self) -> ForestDataReader:
        return ForestDataReader(self)



import pathlib
import lxml.etree as Etree


ODM_SCHEMA_XSD = pathlib.Path().joinpath(
    pathlib.Path(__file__).absolute().parent,
    "../resources", "odm", "schema")


def xsd(file: str):
    def __wrapper__(func):
        def __inner__(cls, xml_file):
            xsd_file = pathlib.Path().joinpath(ODM_SCHEMA_XSD, file)
            try:
                xml_doc = Etree.parse(xml_file)
                schema_xsd = Etree.XMLSchema(file=xsd_file)
                schema_xsd.assertValid(xml_doc)
            except Etree.DocumentInvalid as e:
                raise e
            return True

        return __inner__

    return __wrapper__


class Validator:

    @classmethod
    @xsd(file="xml.xsd")
    def xml(cls, xml_file: str): ...

    @classmethod
    @xsd(file="ODM.xsd")
    def odm(cls, xml_file: str): ...

    @classmethod
    @xsd(file="ODM-admindata.xsd")
    def admindata(cls, xml_file: str): ...

    @classmethod
    @xsd(file="ODM-clinicaldata.xsd")
    def clinicaldata(cls, xml_file: str): ...



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Association(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Association
    """
    
    StudyOID = Model.Attribute()
    
    MetaDataVersionOID = Model.Attribute()

    # 手动修改，因为标准定义了2个相同的子元素
    KeySet = Model.ManyElements()

    Annotation = Model.OneElement()
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class AdminData(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/AdminData+Element
    """
    
    StudyOID = Model.Attribute()
    
    
    User = Model.ManyElements()
    
    Organization = Model.ManyElements()
    
    Location = Model.ManyElements()
    
    SignatureDef = Model.ManyElements()
    



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Query(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Query
    """
    
    OID = Model.Attribute()
    
    Source = Model.Attribute()
    
    Target = Model.Attribute()
    
    Type = Model.Attribute()
    
    State = Model.Attribute()
    
    LastUpdateDatetime = Model.Attribute()
    
    Name = Model.Attribute()
    
    
    Value = Model.OneElement()
    
    AuditRecord = Model.ManyElements()
    



from lxml import etree

from pyodm.core.support.abstract_data_loader import AbstractDataLoader
from pyodm.core.support.abstract_data_reader import AbstractDataReader
from pyodm.factory.cdisc_registry import CdiscRegistry
from pyodm.model.definition import OneElement, ManyElements
from pyodm.utils.etree_utils import EtreeUtils


class XMLDataReader(AbstractDataReader):
    """
    读取 XML 文件的实现
    """

    def __init__(self, registry: CdiscRegistry):
        super().__init__(registry)

    def _new_cdisc_node(self, cdisc_name):
        new_obj = self.registry.get(cdisc_name)()
        return new_obj

    def _set_attribute(self, element, cdisc_node):
        for key, val in element.attrib.items():
            attr = getattr(cdisc_node, key)
            attr.set_name(key)
            attr.set_value(val)

    def _sub_cdisc_node(self, cdisc_node, sub_cdisc_name):
        return getattr(cdisc_node, sub_cdisc_name)

    def _set_text(self, element, cdisc_node):
        if element.text and not element.text.isspace() and element.text.strip():
            cdisc_node.set_value(element.text)

    def _parse(self, element, cdisc_name):
        cdisc_node = self._new_cdisc_node(cdisc_name)
        cdisc_node.set_name(cdisc_name)
        # 处理属性节点
        self._set_attribute(element, cdisc_node)
        # 处理文本节点
        self._set_text(element, cdisc_node)
        for sub_element in element:
            if isinstance(sub_element, etree._Comment): continue  # 忽略注释
            sub_cdisc_name = EtreeUtils.localname(sub_element)
            # 获取 子节点 实例
            sub_cdisc_node = self._sub_cdisc_node(cdisc_node, sub_cdisc_name)
            # 处理类型为 OneElement的Element的对象
            self._one_element(cdisc_node, sub_cdisc_node, sub_element, sub_cdisc_name)
            # 处理类型为 ManyElements的Element的对象
            self._many_element(cdisc_node, sub_cdisc_node, sub_element, sub_cdisc_name)
        return cdisc_node

    def _many_element(self, cdisc_node, sub_cdisc_node, sub_element, sub_cdisc_name):
        if isinstance(sub_cdisc_node, ManyElements):
            many_element = sub_cdisc_node
            many_element.set_name(sub_cdisc_name)
            many_element << self._parse(sub_element, sub_cdisc_name)

    def _one_element(self, cdisc_node, sub_cdisc_node, sub_element, sub_cdisc_name):
        if isinstance(sub_cdisc_node, OneElement):
            one_element = self._parse(sub_element, sub_cdisc_name)
            one_element.set_name(sub_cdisc_name)
            setattr(cdisc_node, sub_cdisc_name, one_element)

    def read(self, loader: AbstractDataLoader):
        tree = loader.load()
        return self._parse(tree.getroot(), EtreeUtils.localname(tree.getroot()))



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class ItemGroupRef(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/ItemGroupRef
    """
    
    ItemGroupOID = Model.Attribute()
    
    MethodOID = Model.Attribute()
    
    OrderNumber = Model.Attribute()
    
    Mandatory = Model.Attribute()
    
    CollectionExceptionConditionOID = Model.Attribute()
    
    



from pyodm.model.definition.cdisc_model import CdiscModel


class Attribute(CdiscModel):
    """
    ODM 标准中对应的 Element 的 Attribute属性
    """

    def __init__(self):
        super().__init__()
        self._value = ""

    def get_name(self): return self._name

    def set_name(self, name): self._name = name

    def get_value(self): return self._value

    def set_value(self, value): self._value = value



from pyodm.core.support.abstract_data_source import AbstractDataSource
from pyodm.core.support.abstract_data_loader import AbstractDataLoader
from pyodm.utils.forest import Forest


class ForestDataLoader(AbstractDataLoader):

    def __init__(self, data_source: AbstractDataSource):
        self._data_source = data_source

    def load(self):
        return Forest.transform(self._data_source.read())



from pyodm.model.definition.cdisc_model import CdiscModel
from pyodm.model.meta.cdisc_odm_entity import CdiscODMEntity
from pyodm.model.definition import Attribute, OneElement, ManyElements
from pyodm.utils.class_utils import ClassUtils


class OdmUtils:
    @staticmethod
    def append(target, odm_element):
        target.array.append(odm_element)
        return target

    @staticmethod
    def merge(base, merged_entity):

        for x_name, x_element in vars(base).items():

            entity_element = ClassUtils.get(merged_entity, x_name)
            if x_name == '_value' and entity_element is not None:
                base.set_value(ClassUtils.get(merged_entity, x_name))

            if not OdmUtils.is_cdisc_model(entity_element): continue
            if entity_element is None: continue
            # if entity_element is None or entity_element.no_use(): continue
            if OdmUtils.is_attribute(x_element) and OdmUtils.is_attribute(entity_element):
                if not entity_element.no_use() and x_element.no_use():
                    x_element.set_value(entity_element.get_value())
                    x_element.set_name(x_name)
            elif OdmUtils.is_one_element(x_element):
                #  如果是 OneElement 的话，则表示该Element 并末使用，跳过该Element
                entity_element = ClassUtils.get(merged_entity, x_name)
                if entity_element is not None and not OdmUtils.is_one_element(entity_element):
                    setattr(base, x_name, entity_element)

            elif OdmUtils.is_many_elements(x_element) and OdmUtils.is_many_elements(entity_element):
                if not x_element.no_use():
                    length = len(x_element.array)
                    for e in entity_element.array:
                        is_like = False
                        for i in range(length):
                            if OdmUtils.like(x_element.array[i], e):
                                OdmUtils.merge(x_element.array[i], e)
                                is_like = True
                                continue
                        if not is_like:
                            OdmUtils.append(x_element, e)
                else:
                    for e in entity_element.array:
                        OdmUtils.append(x_element, e)
                    x_element.set_name(x_name)

            elif OdmUtils.is_entity(x_element) and OdmUtils.is_entity(entity_element):
                # 实例化了 ODM 对象
                if x_element.get_value() != entity_element.get_value():
                    x_element.set_value(entity_element.get_value())
                OdmUtils.merge(x_element, entity_element)
        return base

    @staticmethod
    def like(target, odm_element):
        """
        判断 odm_element 与 target 是否属性相似
        target的属性为 odm_element 的子集　或 odm_element 的属性为 target子集 返回　True, 否则 False
        举例
        target 属性 A B, odm_element 属性 A B 返回 True
        target 属性 A B, odm_element 属性 A B C D 返回 True
        target 属性 A D, odm_element 属性 A B 返回 False
        target 属性 A,B, odm_element 属性 A 返回 True

        :param target:
        :type target:
        :param odm_element:
        :type odm_element:
        :return:
        :rtype:
        """
        target_attributes = list(filter(lambda e: OdmUtils.is_attribute(e) and not e.no_use(), vars(target).values()))
        odm_attributes = list(filter(lambda e: OdmUtils.is_attribute(e) and not e.no_use(), vars(odm_element).values()))
        # 如果 Element 都 没有属性
        if not target_attributes and not odm_attributes: return True
        if not target_attributes: return False
        for attribute in target_attributes:
            if OdmUtils.useless(odm_element, attribute.get_name()): continue
            if attribute.get_value() != odm_element.__dict__.get(attribute.get_name()).get_value():
                return False
        return True

    @staticmethod
    def is_attribute(element):
        return isinstance(element, Attribute)

    @staticmethod
    def is_one_element(element):
        return isinstance(element, OneElement)

    @staticmethod
    def is_many_elements(element):
        return isinstance(element, ManyElements)

    @staticmethod
    def is_entity(element):
        return isinstance(element, CdiscODMEntity)

    @staticmethod
    def is_cdisc_model(element):
        return isinstance(element, CdiscModel) or OdmUtils.is_entity(element)

    @staticmethod
    def useless(element, sub_name):
        """
        判断 sub_name 在 element中是否存在，或没有使用
        :param element:
        :type element:
        :param sub_name:
        :type sub_name:
        :return:
        :rtype:
        """
        return element.__dict__.get(sub_name) is None or element.__dict__.get(sub_name).no_use()

    @staticmethod
    def add(base, element_entity):
        setattr(base, element_entity.get_name(), element_entity)



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class ExternalCodeLib(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/ExternalCodeLib
    """
    
    Library = Model.Attribute()
    
    Method = Model.Attribute()
    
    Version = Model.Attribute()
    
    ref = Model.Attribute()
    
    href = Model.Attribute()
    
    



from pyodm.model.definition import Attribute, OneElement, ManyElements
from pyodm.utils.dict_utils import DictUtils


class ClassUtils:

    @staticmethod
    def filter(clazz, lambda_function) -> dict:
        return DictUtils.filter(vars(clazz), lambda_function)

        # return {key: value for key, value in vars(clazz).items() if lambda_function(key, value)}

    @staticmethod
    def attributes_name(clazz) -> list[str]:
        return list(ClassUtils.filter(clazz, lambda key, value: isinstance(value, Attribute)).keys())


    @staticmethod
    def elements_name(clazz) -> list[str]:
        def condition(key, value):
            return isinstance(value, OneElement) or isinstance(value, ManyElements)

        return list(ClassUtils.filter(clazz, condition).keys())

    @staticmethod
    def get(entity, key: str):
        return entity.__dict__.get(key)



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class Signature(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/Signature+Element
    """
    
    ID = Model.Attribute()
    
    
    UserRef = Model.OneElement()
    
    LocationRef = Model.OneElement()
    
    SignatureRef = Model.OneElement()
    
    DateTimeStamp = Model.OneElement()
    




import pytest

from pyodm.core.source.path_source import PathSource
from pyodm.factory.cdsic_xml_xsd_factory import CdiscXMLXsdFactory
from pyodm.tests.test_cdisc_factory.common_assert import assert_case1, assert_case2, \
    assert_case3, assert_case4, assert_case5, assert_case6
from pyodm.utils.path_utils import PathUtils


@pytest.fixture
def xsd_files():
    base = PathUtils.folder("pyodm", __file__).joinpath("model", "v2", "resources", "schema")
    xsds = ["ODM-admindata.xsd", "ODM-clinicaldata.xsd", "ODM-foundation.xsd",
            "ODM-protocol.xsd", "ODM-referencedata.xsd", "ODM-study.xsd"]
    return [base.joinpath(xsd) for xsd in xsds]


def test_case1(xsd_files):
    xsd_file = [xsd_files[0]]
    data_file = PathSource("resources", "data", "user.xml")
    cdisc = CdiscXMLXsdFactory(data_file=data_file, xsd_files=xsd_file)
    assert_case1(cdisc)


def test_case2(xsd_files):
    data_file = PathSource("resources", "data", "odm_template.xml")
    cdisc = CdiscXMLXsdFactory(data_file=data_file, xsd_files=xsd_files)
    assert_case2(cdisc)


def test_case3(xsd_files):
    data_file = PathSource("resources", "data", "test_case2.xml")
    cdisc = CdiscXMLXsdFactory(data_file=data_file, xsd_files=xsd_files)
    assert_case3(cdisc)


def test_case4(xsd_files):
    data_file = PathSource("resources", "data", "transaction_history.xml")
    cdisc = CdiscXMLXsdFactory(data_file=data_file, xsd_files=xsd_files)
    assert_case4(cdisc)


def test_case5(xsd_files):
    data_file = PathSource("resources", "data", "transaction_history.xml")
    cdisc = CdiscXMLXsdFactory(data_file=data_file, xsd_files=xsd_files)
    assert_case5(cdisc)


def test_case6(xsd_files):
    data_file = PathSource("resources", "data", "odm_testcase5.xml")
    cdisc = CdiscXMLXsdFactory(data_file=data_file, xsd_files=xsd_files)
    assert_case6(cdisc)



import pyodm.model.definition as Model
import pyodm.model.meta.cdisc_odm_entity as Meta


class StudyStructure(Meta.CdiscODMEntity):
    """
    https://wiki.cdisc.org/display/ODM2/StudyStructure
    """
    
    
    Description = Model.OneElement()
    
    Arm = Model.ManyElements()
    
    Epoch = Model.ManyElements()
    
    WorkflowRef = Model.OneElement()
    



import pathlib

from pyodm.core.xml.writer.cdisc_xml_writer import CdiscXmlWriter


def assert_case1(cdisc):
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()


def assert_case2(cdisc):
    x = cdisc.odm().ClinicalData.first().SubjectData.first().StudyEventData.first().ItemGroupData.first().ItemData.first().Value
    assert x.first().get_value() == '2200-10-01'
    query = cdisc.odm().ClinicalData.first().SubjectData.first().StudyEventData.first().ItemGroupData.first().ItemData.find().Query.find()
    assert query.Value.get_value() == 'Value 2200-10-01 is in the future, please correct'
    assert query.AuditRecord.first().DateTimeStamp.get_value() == '2021-03-10T13:36:51.668-00:00'
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()


def assert_case3(cdisc):
    x = cdisc.odm().ClinicalData
    assert x.first().StudyOID.get_value() == '11'
    assert x.index(1).StudyOID.get_value() == "EX001"
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()


def assert_case4(cdisc):
    subject = cdisc.odm().ClinicalData.first().SubjectData.first()
    assert subject.SubjectKey.get_value() == '1001'
    query = subject.StudyEventData.find().ItemGroupData.first().ItemData.first().Query._array
    assert query[0].Value.get_value() == "Value is in the future, please correct"
    for i in query[0].AuditRecord._array:
        print(i.ReasonForChange.get_value())
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()


def assert_case5(cdisc):
    cdisc.odm().ClinicalData.usage(lambda x: print(x.first().StudyOID.get_value()))


def assert_case6(cdisc):
    subject = cdisc.odm().ClinicalData.first().SubjectData.first()
    item_group_data = subject.StudyEventData.first().ItemGroupData.find()
    item_group_data1 = item_group_data.ItemGroupData.index(0)
    assert item_group_data1.ItemGroupOID.get_value() == "IG.WBC"
    item_data = item_group_data1.ItemData.find()
    assert item_data.Value.first().get_value() == '5.2'
    item_data = item_group_data1.ItemData.index(1)
    assert item_data.Value.first().get_value() == '10*3/uL'
    item_group_data2 = item_group_data.ItemGroupData.find()
    assert item_group_data2.ItemGroupOID.get_value() == "IT.RBC"
    item_data = item_group_data2.ItemData.first()
    assert item_data.IsNull.get_value() == "Yes"
    cw = CdiscXmlWriter(cdisc.odm(), out_put=pathlib.Path("test.xml"))
    cw.write()
